#  Copyright (c) 2022 by Amplo.

from amplo.documenting.binary import BinaryDocumenting
from amplo.documenting.multiclass import MultiDocumenting
from amplo.documenting.regression import RegressionDocumenting
